"""Simplified testing for Python functions"""

from typing import Callable


def test(function: Callable, test_table: list) -> None:
    """Test the function with the test_table. Report failed tests.

    Args:
        function (Callable): The function to be tested.
        test_table (list): The list of tests. Each element of test_table is a list or
            tuple with: a string (the test case name); one or more values (the inputs to the function);
            the expected output value.
    """
    not_tested = f"The test table is invalid, {function.__name__} has NOT been tested."
    if not isinstance(test_table, (list, tuple)):
        print(f"Error: The test table must be a list or tuple.\n{not_tested}")
        return
    # The number of arguments function expects.
    expects = function.__code__.co_argcount
    invalid = []
    for num, test_case in enumerate(test_table, start=1):
        if not isinstance(test_case, (list, tuple)):
            invalid.append(f"test case {num} must be a list or tuple.")
        elif not isinstance(test_case[0], str):
            invalid.append(f"test case {num} must have string as first element.")
        elif len(test_case[1:-1]) != expects:
            num_args = len(test_case[1:-1])
            invalid.append(
                f'test case "{test_case[0]}" has {num_args} input(s) instead of {expects}'
            )
    if invalid:
        for msg in invalid:
            print(f"Error: {msg}")
        print(not_tested)
        return
    else:
        print(f"Testing {function.__name__}...")
        passed = failed = 0
        for test_case in test_table:
            name = test_case[0]
            inputs = test_case[1:-1]
            expected = test_case[-1]
            actual = function(*inputs)
            if actual != expected:
                print(name, "FAILED:", actual, "instead of", expected)
                failed += 1
            else:
                passed += 1
        print("Tests finished:", passed, "passed,", failed, "failed.")



"""Tools for measuring and plotting run-times"""

from typing import Callable

import timeit
import matplotlib.pyplot as plt


# Input generators
# ----------------


def int_value(n: int) -> tuple[int]:
    """Return the value n."""
    return (n,)


# Timing functions
# -----------------


def scale_and_unit(seconds: float) -> tuple[int, str]:
    """Given a time in seconds, return the appropriate scale factor and time unit."""
    if seconds < 1e-6:
        return 1_000_000_000, "ns"
    if seconds < 1e-3:
        return 1_000_000, "µs"
    if seconds < 1:
        return 1_000, "ms"
    return 1, "s"


def time_it(function: Callable, *args, loops=0, repeat=3) -> float:
    """Return the fastest time, in seconds, of running `function` on `*args`.

    By default (zero loops), use enough loops to take >= 0.2 seconds.
    """
    assert loops >= 0, "loops must be non-negative"
    assert repeat > 0, "repeat must be positive"

    timer = timeit.Timer(lambda: function(*args))
    if loops == 0:
        loops, run_time = timer.autorange()
        run_times = timer.repeat(repeat - 1, loops)
        run_times.append(run_time)
    else:
        run_times = timer.repeat(repeat, loops)
    return min(run_times) / loops


def time_cases(
    function: Callable,
    cases: list[Callable],
    start: int,
    double: int,
    text: bool = True,
    chart: bool = False,
) -> None:
    """Print or plot the run-times of `function` for different input cases.
    
    `time_cases` prints or plots the run-times of a single function using a list of 
    different input generators. Inputs are generated based on a starting size and are 
    doubled a specified number of times.

    Args:
        function (Callable): A function whose run-times will be measured.
        cases (list[Callable]): A list of 1 to 6 functions to generate inputs of different cases,
            e.g. best-, normal- and worst-case.
        start (int): The starting size for the inputs. Must be positive.
        double (int): The number of times to double the input size. Must be non-negative.
        text (bool, optional): If True, print the run-times in text format.
        chart (bool, optional): If True, plot the run-times using a chart.
    
    Raises:
        AssertionError: If input conditions are not satisfied.
    """
    assert start > 0, "the start size must be positive"
    assert double >= 0, "must double the input size at least zero times"
    assert 0 < len(cases) < 7, "there must be 1 to 6 input functions"
    assert text or chart, "at least one of text and chart must be enabled"

    sizes = [start]  # the input sizes used
    for _ in range(double):
        sizes.append(sizes[-1] * 2)
    scale = unit = None  # no scale determined yet
    if chart:
        markers = ["bo-", "ko--", "ro:", "ys-", "cs--", "gs:"]
        times: list[list[float]] = []
        for _ in range(len(cases)):
            times.append([])
        plt.title(f"Run-times for {function.__name__}")
    if text:
        print(f"Run-times for {function.__name__}\n")
        print("Input size", end=" ")
        for case in cases:
            print(f"{case.__name__[:15]:>15}", end=" ")
    for size in sizes:
        if text:
            print(f"\n{size:>10}", end=" ")
        for index in range(len(cases)):
            instance = cases[index](size)
            run_time = time_it(function, *instance)
            if not scale:
                scale, unit = scale_and_unit(run_time)
            run_time = run_time * scale
            if chart:
                times[index].append(run_time)
            if text:
                print(f"{run_time:>15.1f}", end=" ")
        if text:
            print(f"{unit}", end="")
    if chart:
        plt.xlabel("Input size")
        plt.ylabel(f"Run-time ({unit})")
        for index in range(len(cases)):
            plt.plot(sizes, times[index], markers[index], label=cases[index].__name__)
        plt.legend()
        plt.show()


def time_functions(
    functions: list[Callable],
    inputs: Callable,
    start: int,
    double: int,
    text: bool = True,
    chart: bool = False,
    value: bool = False,
) -> None:
    """Print or plot the run-times of different functions for the same inputs.

    `time_functions` prints or plots the run-times given list of functions and an input 
    generator. Inputs are generated based on a starting size and are doubled a specified 
    number of times.

    Args:
        functions (list[Callable]): A list of functions whose run-times will be measured.
            Must be 1 to 6 functions.
        inputs (Callable): A function to generate inputs when given a specific size.
        start (int): The starting size for the inputs. Must be positive.
        double (int): The number of times to double the input size. Must be non-negative.
        text (bool, optional): If True, print the run-times in text format
        chart (bool, optional): If True plot the run-times using a chart.
        value (bool, optional): If True x-axis is labelled "Input value" otherwise "Input size".

    Raises:
        AssertionError: If input conditions are not satisfied.
    """
    assert start > 0, "the start size/value can't be negative"
    assert double >= 0, "must double the input size/value at least zero times"
    assert 0 < len(functions) < 7, "there must be 1 to 6 functions"
    assert text or chart, "at least one of text and chart must be enabled"

    x_label = "Input " + ("value" if value else "size")
    text_width = len(x_label)
    sizes = [start]  # the input sizes used
    for _ in range(double):
        sizes.append(sizes[-1] * 2)
    scale = unit = None  # no scale determined yet
    if chart:
        markers = ["bo-", "ko--", "ro:", "ys-", "cs--", "gs:"]
        times: list[list[float]] = []
        for _ in range(len(functions)):
            times.append([])
        plt.title(f"Inputs generated by {inputs.__name__}")
    if text:
        print(f"Inputs generated by {inputs.__name__}\n")
        print(x_label, end=" ")
        for index in range(len(functions)):
            print(f"{functions[index].__name__[:15]:>15}", end=" ")
    for size in sizes:
        if text:
            print(f"\n{size:>{text_width}}", end=" ")
        instance = inputs(size)
        for index in range(len(functions)):
            run_time = time_it(functions[index], *instance)
            if not scale:
                scale, unit = scale_and_unit(run_time)
            run_time = run_time * scale
            if chart:
                times[index].append(run_time)
            if text:
                print(f"{run_time:>15.1f}", end=" ")
        if text:
            print(f"{unit}", end="")
    if chart:
        plt.xlabel(x_label)
        plt.ylabel(f"Run-time ({unit})")
        for index in range(len(functions)):
            plt.plot(
                sizes, times[index], markers[index], label=functions[index].__name__
            )
        plt.legend()
        plt.show()


def time_functions_int(
    functions: list[Callable],
    generator: Callable = int_value,
    start: int = 1,
    double: int = 10,
    text: bool = True,
    chart: bool = True,
) -> None:
    """Time functions that take a single integer as input.
    
    `time_functions_int` uses `time_functions` to measure and display the run-times
    of a given list of functions that accept a single integer input. The integer inputs
    are generated starting from a specified value that defaults to 1, and are doubled
    a specified number of times that defaults to 10.

    Args:
        functions (list[Callable]): A list of functions whose run-times will be measured.
            Each function must accept a single integer argument. Must be 1 to 6 functions.
        generator (Callable, optional): A function to generate integer inputs. Defaults to
            `int_value`, which returns a tuple containing the input integer.
        start (int, optional): The starting integer value for inputs. Defaults to 1.
            Must be positive.
        double (int, optional): The number of times to double the input integer value.
            Defaults to 10. Must be non-negative.
        text (bool, optional): If True, print the run-times in text format. 
        chart (bool, optional): If True, plot the run-times using a chart. 
    """
    time_functions(functions, generator, start, double, text, chart, True)



"""Linting tools for Jupyter Notebook environments"""

import argparse
import json
import os
import re
import subprocess
import tempfile
from typing import Callable
from subprocess import CompletedProcess

from IPython.core.inputtransformer2 import TransformerManager
from IPython.core.magic import register_line_magic
from IPython.display import display_markdown


def show_allowed_errors(checker: str, output: CompletedProcess, filename: str) -> None:
    """Print the errors for the given file in allowed's output."""
    if output.returncode > 0:
        display_markdown(f"**{checker}** didn't check code:", raw=True)
        print(output.stderr if output.stderr else output.stdout)
    else:
        md = [f"**{checker}** found issues:"]
        for line in output.stdout.split("\n"):
            if "syntax" in line.lower() and "error" in line.lower():
                continue  # syntax errors already reported when running the cell
            if m := re.match(rf".*{filename}[^\d]*(\d+[^:]*:.*)", line):
                md.append(f"- {m.group(1)}")
        if len(md) > 1:
            display_markdown("\n".join(md), raw=True)


def show_ruff_json(checker: str, output: CompletedProcess, filename: str) -> None:
    """Print the errors in ruff's JSON output for the given file."""
    if output.stderr:
        # ignore syntax errors: they're reported on running the cell
        if "Failed to parse" in output.stderr:
            return
        text = "has warning:" if "warning" in output.stderr.lower() else "didn't check code:"
        display_markdown(f"**{checker}** {text}", raw=True)
        print(output.stderr)
    if errors := json.loads(output.stdout):
        md = [f"**{checker}** found issues:"]
        # the following assumes errors come in line order
        for error in errors:
            line = error["location"]["row"]
            code = error["code"]
            url = error["url"]
            msg = error["message"]
            if error["fix"]:
                msg += f". Suggested fix: {error['fix']['message']}"
            md.append(rf"- {line}: \[[{code}]({url})\] {msg}")
        display_markdown("\n".join(md), raw=True)


def show_pytype_errors(checker: str, output: CompletedProcess, filename: str) -> None:
    """Print the errors in pytype's output for the given file."""
    if output.stderr:
        text = "has warning:" if "warning" in output.stderr.lower() else "didn't check code:"
        display_markdown(f"**{checker}** {text}", raw=True)
        print(output.stderr)
    md = [f"**{checker}** found issues:"]
    for error in output.stdout.split("\n"):
        if "syntax" in error.lower() and "error" in error.lower():
            continue  # syntax errors already reported when running the cell
        if "python-compiler-error" in error:
            continue
        if m := re.match(rf".*{filename}[^\d]*(\d+)[^:]*:(.*)\[(.*)\]", error):
            line = m.group(1)
            msg = m.group(2)
            code = m.group(3)
            md.append(
                rf"- {line}:{msg}\[[{code}](https://google.github.io/pytype/errors.html#{code})\]"
            )
    if len(md) > 1:
        display_markdown("\n".join(md), raw=True)

# register the supported checkers, their commands and the output processor
checkers: dict[str, tuple[str, Callable]] = {
    "pytype": [["pytype"], show_pytype_errors],
    "allowed": [["allowed"], show_allowed_errors],
    "ruff": [["ruff", "check", "--output-format", "json"], show_ruff_json],
}
# initially no checker is active
active: set[str] = set()


def process_status(name: str, status: str) -> None:
    """Process the status of a checker."""
    if status is None:
        print(name, "is", "active" if name in active else "inactive")
    elif status == "on":
        active.add(name)
        print(name, "was activated")
    elif status == "off":
        active.discard(name)
        print(name, "was deactivated")


@register_line_magic
def pytype(line: str) -> None:
    """Activate/deactivate the [pytype linter](https://google.github.io/pytype).

    When active, the linter checks each code cell that is executed for type errors.

    - `%pytype on ...` activates the linter with the command options given after `on`
    - `%pytype on` is equal to `%pytype on --disable name-error,import-error`
    - `%pytype off` deactivates the linter
    - `%pytype` shows the current status of the linter
    - `%pytype?` shows this documentation and the command's options

    For a list of possible options `...`, enter `!pytype -h` in a code cell.
    Some options may not be appropriate when running pytype within a notebook.

    The `--disable` option expects a list of
    [errors](https://google.github.io/pytype/errors.html) to ignore, without spaces.
    By default, `%pylint` ignores these errors:

    - name-error: cells often use names defined in previous cells
    - import-error: pytype doesn't find local modules
    """
    parser = argparse.ArgumentParser("pytype")
    parser.add_argument("status",
        choices=["on", "off"],
        type=str.lower,
        help="Activate or deactivate the linter. If omitted, show the current status.",
        nargs="?",
        default=None,
    )
    parser.add_argument("-d", "--disable",
        default="name-error,import-error",
        help="Comma or space-separated list of error names to ignore",
    )
    known, unknown = parser.parse_known_args(line.split())
    if known.status != "on" and (known or unknown):
        print("warning: ignoring additional options for %pytype")
    else:
        checkers["pytype"][0] = ["pytype", "--disable", known.disable] + unknown
    process_status("pytype", known.status)


@register_line_magic
def allowed(line: str) -> None:
    """Activate/deactivate the [allowed linter](https://dsa-ou.github.io/allowed).

    When active, the linter checks each code cell that is executed for any
    Python constructs that are not listed in the given configuration file.

    - `%allowed on ...` activates the linter with any command options given after `on`
    - `%allowed on` is equal to `%allowed on --config m269.json`
    - `%allowed off` deactivates the linter
    - `%allowed` shows the current status of the linter
    - `%allowed?` shows this documentation and the command's options

    For a list of possible options `...`, enter `!allowed -h` in a code cell.
    Some options may not be appropriate when running `allowed` within a notebook.

    The `--config` option expects `m269.json`, `tm112.json` or the name of a JSON file
    with your own [configuration](https://dsa-ou.github.io/allowed/docs/configuration.html).
    """
    parser = argparse.ArgumentParser("allowed")
    parser.add_argument("status",
        choices=["on", "off"],
        type=str.lower,
        help="Activate or deactivate the linter. If omitted, show the current status.",
        nargs="?",
        default=None,
    )
    parser.add_argument("-c", "--config",
        default=None,
        help="Use configuration file CONFIG (default: m269.json).",
    )
    known, unknown = parser.parse_known_args(line.split())
    if known.status != "on" and (known or unknown):
        print("warning: ignoring additional options for %allowed")
    else:
        config = ["-c", known.config] if known.config else []
        checkers["allowed"][0] = ["allowed"] + config + unknown
    process_status("allowed", known.status)


@register_line_magic
def ruff(line: str) -> None:
    """Activate/deactivate the [Ruff linter](https://docs.astral.sh/ruff).

    When active, the linter checks each code cell that is executed
    against the selected code style rules.

    - `%ruff on ...` activates the linter with any command options given after `on`
      (see [ruff's list of rules])
    - `%ruff on` is equal to `%ruff on --select A,B,C90,D,E,W,F,N,PL --ignore D100,W292,F401,F821,D203,D213,D415`
    - `%ruff off` deactivates the linter
    - `%ruff` shows the current status of the linter
    - `%ruff?` shows this documentation

    The command `%ruff on ...` will run `ruff check --output-format json ...` on each cell.
    For a list of the possible options `...`, enter `!ruff help check` in a code cell.
    Some options may not be appropriate when running Ruff within a notebook.

    The `--select` and `--ignore` options expect a list
    of [rule codes](https://docs.astral.sh/ruff/rules), without spaces.

    By default, `%ruff` enables the rules from
    flake8-builtins (A), flake8-bugbear (B), mccabe (C90), pydocstyle (D),
    pycodestyle (E, W), pyflakes (F), pep8-naming (N), and pylint (PL).

    By default, `%ruff` ignores these rules:

    - D100: Missing docstring in public module (notebooks aren't modules)
    - W292: No newline at end of file (cells don't have a trailing blank line)
    - F401: Imported but unused (cells often import modules that are used later)
    - F821: Undefined name (cells often use names defined in previous cells)
    - D203: use D211 instead (no blank line before class docstring)
    - D213: use D212 instead (docstring starts right after \"\"\")
    - D415: use D400 instead (first line of docstring must end in .)
    """
    parser = argparse.ArgumentParser("ruff")
    parser.add_argument("status",
        choices=["on", "off"],
        type=str.lower,
        help="Activate or deactivate the linter. If omitted, show the current status.",
        nargs="?",
        default=None,
    )
    parser.add_argument("--select",
        help="Comma-separated list of rule codes to enable",
        type=str,
        default="A,B,C90,D,E,W,F,N,PL",
    )
    parser.add_argument("--ignore",
        help="Comma-separated list of rule codes to ignore",
        type=str,
        default="D100,W292,F401,F821,D203,D213,D415",
    )
    known, unknown = parser.parse_known_args(line.split())
    if known.status != "on" and (known or unknown):
        print("warning: ignoring additional options for %ruff")
    else:
        base = ["ruff", "check", "--output-format", "json"]
        rules = ["--select", known.select, "--ignore", known.ignore]
        checkers["ruff"][0] = base + rules + unknown
    process_status("ruff", known.status)


def run_checkers(result) -> None:
    """Run all active checkers after a cell is executed."""
    if not active:
        return
    try:
        with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as temp:
            # transform IPython to pure Python to avoid linters reporting syntax errors
            temp.write(TransformerManager().transform_cell(result.info.raw_cell))
            # Handle Windows file paths
            temp_name = temp.name.replace("\\", "/")
        for checker in active:
            command, display = checkers[checker]
            lint_file = command + [temp_name]
            try:
                output = subprocess.run(
                    lint_file, capture_output=True, text=True, check=False,
                )
                display(checker, output, temp_name)
            except Exception as e:
                print(f"Error on executing {command[0]}:\n{e}")
    except Exception as e:
        print(f"Error on writing cell to a temporary file:\n{e}")
    finally:
        os.remove(temp_name)


def load_ipython_extension(ipython):
    """Load the ipython extension, and register run_checkers with post_cell_run

    This function hooks into the ipython extension system so the magic commands defined
    in this module can be loaded with `load_ext algoesup.magics`. It also registers
    `run_checkers` with the `post_run_cell` event so the linters are run with the
    contents of each ipython cell after it has been executed.
    """
    ipython.events.register("post_run_cell", run_checkers)  # type: ignore[name-defined]
